""" Python Obit inteferometer (UV) data class

This class contains interoferometric data and allows access.
An ObitUV is the front end to a persistent disk resident structure.
There maybe (usually are) associated tables which either describe
the data or contain calibration and/or editing information.
Both FITS (as Tables) and AIPS cataloged data are supported.
Most access to UV data is through functions as the volume of the data is
inappropriate to be processed directly in python.

UV Members with python interfaces:
exist - True if object previously existed prior to object creation
InfoList  - used to pass instructions to processing
UVDesc - Astronomical labeling of the data

Additional Functions are available in UVUtil.
"""
# $Id: UV.py,v 1.47 2007/04/16 16:46:11 bcotton Exp $
#-----------------------------------------------------------------------
#  Copyright (C) 2004-2007
#  Associated Universities, Inc. Washington DC, USA.
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public
#  License along with this program; if not, write to the Free
#  Software Foundation, Inc., 675 Massachusetts Ave, Cambridge,
#  MA 02139, USA.
#
#  Correspondence concerning this software should be addressed as follows:
#         Internet email: bcotton@nrao.edu.
#         Postal address: William Cotton
#                         National Radio Astronomy Observatory
#                         520 Edgemont Road
#                         Charlottesville, VA 22903-2475 USA
#-----------------------------------------------------------------------
 
# Python shadow class to ObitUV class
import OErr, InfoList, Table, AIPSDir, FITSDir, OSystem
import Obit, TableList,  UVDesc, string

class UVPtr :
    def __init__(self,this):
        self.this = this
    def __setattr__(self,name,value):
        if name == "me" :
            # Out with the old
            Obit.UVUnref(Obit.UV_me_get(self.this))
            # In with the new
            Obit.UV_me_set(self.this,value)
            return
        self.__dict__[name] = value
    def __getattr__(self,name):
        if self.__class__ != UV:
            return
        if name == "me" : 
            return Obit.UV_me_get(self.this)
        # Functions to return members
        if name=="List":
            return PGetList(self)
        if name=="TableList":
            return PGetTableList(self)
        if name=="Desc":
            return PGetDesc(self)
        if name=="VisBuf":
            return PGetVisBuf(self)
        raise AttributeError,str(name)  # Unknown
    def __repr__(self):
        if self.__class__ != UV:
            return
        return "<C UV instance> " + Obit.UVGetName(self.me)
class UV(UVPtr):
    """ Python Obit inteferometer (UV) data class
    
    UV Members with python interfaces:
    InfoList  - used to pass instructions to processing
    UVDesc - Astronomical labeling of the data
    
    Additional Functions are available in UVUtil.
    """
    def __init__(self, name) :
        self.this = Obit.new_UV(name)
    def __del__(self):
        if Obit!=None:
            Obit.delete_UV(self.this)
            
    def Zap (self, err):
        """ Delete underlying files and the basic object.
        
        self      = Python UV object
        err       = Python Obit Error/message stack
        """
        PZap(self, err)
        # end Zap
        
    def Open (self, access, err):
        """ Open a UV data persistent (disk) form
        
        self   = Python UV object
        access    = access READONLY (1), WRITEONLY (2), READWRITE(3)
        err       = Python Obit Error/message stack
        """
        POpen(self, access, err)
        # end Open
        
    def Close (self, err):
        """ Close a UV  persistent (disk) form
        
        self      = Python UV object
        err       = Python Obit Error/message stack
        """
        PClose (self, err)
        # end Close
        
    def NewTable (self, access, tabType, tabVer, err, \
                  numOrb=0, numPCal=3, numIF=1, numPol=1, \
                  numTerm=0, numChan=1, numTones=1, numBand=1, \
                  numTabs=1, npoly=1, numCoef=5):
        """ Return the specified associated table
        
        Table will be created if necessary.
        self      = Python UV object
        access    = access code 1=READONLY, 2=WRITEONLY, 3=READWRITE
        tabType   = Table type, e.g. "AIPS AN"
        tabVer    = table version, if > 0 on input that table returned,
                    if 0 on input, the highest version is used.
        err       = Python Obit Error/message stack
        Optional parameters, values only used if table created
        numOrb    = Number of orbital parameters (AN)
        numPCal   = Number of polarization parameters (AN)
        numIF     = Number of IFs (FQ, SN, CL, BP, BL, TY, CQ)
        numPol    = Number of Stokes' (SN, CL, BP, BL, PC, TY, GC, MC, IM)
        numTerm   = Number of terms in model polynomial (CL)
        numChan   = Number of spectral channels (BP)
        numTomes  = Number of Phase cal tones (PC)
        numTabs   = Number of ??? (GC)
        numCoef   = Number of polynomial coefficents (NI)
        numBand   = Number of Bands(?) (IM, GC)
        npoly     = number of polynomial terms (IM)
        """
        return PGetTable (self, access, tabType, tabVer, err,\
                          numOrb=numOrb, numPCal=numPCal, numIF=numIF, \
                          numPol=numPol, numTerm=numTerm, numChan=numChan,\
                          numTones=numTones, numBand=numBand, \
                          numTabs=numTabs, npoly=npoly, numCoef=numCoef)
        # end NewTable

    def Header (self, err):
        """ Write image header on output
        
        self   = Python Obit UV object
        err    = Python Obit Error/message stack
        """
        PHeader (self, err)
        # end Header
        
    def ZapTable (self, tabType, tabVer, err):
        """ Destroy specified table
        
        self      = Python UV object
        tabType   = Table type, e.g. "AIPS CC"
        tabVer    = table version, integer
        err       = Python Obit Error/message stack
        """
        PZapTable (self, tabType, tabVer, err)
        # end ZapTable

    def UpdateTables (self, err):
        """ Update any disk resident structures about the current tables
        
        self      = Python Image object
        err       = Python Obit Error/message stack
        """
        PUpdateTables (self, err)
        # end UpdateTables
        
    def UpdateDesc (self, err, Desc=None):
        """ Update any disk resident structures about descriptor
        
        self      = Python UV object
        err       = Python Obit Error/message stack
        Desc      = Descriptor, if None then use current descriptor
                    Contents can be accessed throuth the Dict member
        """
        PUpdateDesc (self, err, Desc=Desc)
        # end UpdateDesc
        
    # End of class member functions (i.e. invoked by x.func())

# Symbolic names for access codes
READONLY  = 1
WRITEONLY = 2
READWRITE = 3

def newPFUV(name, filename, disk, exists, err, verbose=True):
    """ Create and initialize an FITS based UV structure
    
    Create, set initial access information (full image, plane at a time)
    and if exists verifies the file.
    Sets buffer to hold 1000 vis.
    Returns the Python UV object
    isOK member set to indicate success
    name     = name desired for object (labeling purposes)
    filename = name of FITS file
    disk     = FITS directory number
    exists   = if true then the file is opened and closed to verify
    err      = Python Obit Error/message stack
    verbose  = If true any give error messages, else suppress
    """
    ################################################################
    out = UV (name)
    out.isOK = True  # until proven otherwise
    # Does it really previously exist?
    out.exist = FITSDir.PExist(disk, filename, err)
    Obit.UVSetFITS(out.me, 1000, disk, filename, err.me)
    if exists:
        Obit.UVfullInstantiate (out.me, 1, err.me)
        if err.isErr:
            OErr.printErrMsg(err, "Error verifying UV data")
    # show any errors if wanted
    if  verbose and err.isErr:
        out.isOK = False
        OErr.printErrMsg(err, "newPFUV: Error verifying file")
    elif err.isErr:
        out.isOK = False
        OErr.PClear(err)  # Clear unwanted messages
    
    # It work?
    if not out.isOK:
        return out
    
    out.FileType = 'FITS'
    out.FileName = filename
    out.Disk     = disk
    out.Otype    = "UV"
    return out
    # end newPFUV
    
    
def newPAUV(name, Aname, Aclass, disk, seq, exists, err, verbose=True):
    """ Create and initialize an AIPS based UV structure
    
    Create, set initial access information (full image, plane at a time)
    and if exists verifies the file.
    Sets buffer to hold 1000 vis.
    Returns the Python UV object
    isOK member set to indicate success
    name     = name desired for object (labeling purposes)
    Aname    = AIPS name of file
    Aclass   = AIPS class of file
    seq      = AIPS sequence number of file
    disk     = FITS directory number
    exists   = if true then the file is opened and closed to verify
    err      = Python Obit Error/message stack
    verbose  = If true any give error messages, else suppress
    """
    ################################################################
    out = UV (name)
    user = OSystem.PGetAIPSuser()
    out.isOK = True  # until proven otherwise
    # Does it really previously exist?
    test = AIPSDir.PTestCNO(disk, user, Aname, Aclass, "UV", seq, err)
    out.exist = test>0
    if exists:
        cno = AIPSDir.PFindCNO(disk, user, Aname, Aclass, "UV", seq, err)
        Obit.UVSetAIPS(out.me, 1000, disk, cno, user, err.me)
        Obit.UVfullInstantiate (out.me, 1, err.me)
    else:
        cno = AIPSDir.PAlloc(disk, user, Aname, Aclass, "UV", seq, err)
        Obit.UVSetAIPS(out.me, 1000, disk, cno, user, err.me)

    # show any errors if wanted
    if verbose and err.isErr:
        out.isOK = False
        OErr.printErrMsg(err, "Error finding AIPS catalog entry")
    elif err.isErr:
        out.isOK = False
        OErr.PClear(err)  # Clear unwanted messages
    
    # It work?
    if not out.isOK:
        return out
    
    out.FileType = 'AIPS'
    out.Disk   = disk
    out.Aname  = Aname
    out.Aclass = Aclass
    out.Aseq   = seq 
    out.Otype  = "UV"
    out.Acno   = cno
    return out
    # end newPAUV

def newPACNO(disk, cno, exists, err, verbose=True):
    """ Create and initialize an AIPS based UV structure

    Create, set initial access information 
    and if exists verifies the file.
    Sets buffer to hold 1000 vis.
    Returns the Python UV object
    isOK member set to indicate success
    disk     = AIPS directory number
    cno      = AIPS catalog number
    exists   = if true then the file is opened and closed to verify
    err      = Python Obit Error/message stack
    verbose  = If true any give error messages, else suppress
    """
    ################################################################
    out = UV ("AIPS UVdata")
    user = OSystem.PGetAIPSuser()
    out.isOK = True  # until proven otherwise
    # print "disk, aseq", disk, seq
    # Does it really previously exist?
    test = AIPSDir.PInfo(disk, user, cno, err)
    out.exist = test != None
    
    if exists:
        Obit.UVSetAIPS(out.me, 1000, disk, cno, user, err.me)
        Obit.UVfullInstantiate (out.me, 1, err.me)

    else:
        Obit.UVSetAIPS(out.me, 1000, disk, cno, user, err.me)

    # show any errors if wanted
    if verbose and err.isErr:
        out.isOK = False
        OErr.printErrMsg(err, "Error finding AIPS catalog entry")
    elif err.isErr:
        out.isOK = False
        OErr.PClear(err)  # Clear unwanted messages

    # It work?
    if not out.isOK:
        return out
    
    # Add File info
    out.FileType = 'AIPS'
    out.Disk   = disk
    out.Acno   = cno
    # Lookup name etc
    s = AIPSDir.PInfo (disk, user, cno, err)
    # parse returned string
    Aname = s[0:12]
    Aclass = s[13:19]
    Aseq = int(s[20:25])
    Atype = s[26:28]
    out.Aname  = Aname
    out.Aclass = Aclass
    out.Aseq   = Aseq 
    out.Otype  = "UV"
    return out   # seems OK
    # end newPACNO

    
def PScratch (inUV, err):
    """ Create a scratch file suitable for accepting the data to be read from inUV
    
    A scratch UV is more or less the same as a normal UV except that it is
    automatically deleted on the final unreference.
    inUV      = Python UV object
    err       = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    outUV    = UV("None")
    outUV.me = Obit.UVScratch (inUV.me, err.me);
    if err.isErr:
        OErr.printErrMsg(err, "Error creating scratch file")
    return outUV
    # end PScratch

def PZap (inUV, err):
    """ Delete underlying files and the basic object.
    
    inUV      = Python UV object
    err       = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    Obit.UVZap (inUV.me, err.me)
    if err.isErr:
        OErr.printErrMsg(err, "Error Zapping UV data")
    # end PZap

def PRename (inUV, err, newFITSName=None, \
             newAIPSName="            ", \
             newAIPSClass="      ", newAIPSSeq=0):
    """ Rename underlying files

    inUV   = Python UV object
    err       = Python Obit Error/message stack
    For FITS files:
    newFITSName = new name for FITS file

    For AIPS:
    newAIPSName  = New AIPS Name (max 12 char) Blank => don't change.
    newAIPSClass = New AIPS Class (max 6 char) Blank => don't change.
    newAIPSSeq   = New AIPS Sequence number, 0 => unique value
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    if len(newAIPSName)>12:
        raise RuntimeError,"New AIPS Name too long"
    if len(newAIPSClass)>6:
        raise RuntimeError,"New AIPS Class too long"
    #
    # Set controls
    inInfo = PGetList(inUV)    # 
    dim = [1,1,1,1,1]
    InfoList.PAlwaysPutInt     (inInfo, "newSeq",       dim, [newAIPSSeq])
    dim[0] = 12
    InfoList.PAlwaysPutString  (inInfo, "newName",    dim, [newAIPSName])
    dim[0] = 6
    InfoList.PAlwaysPutString  (inInfo, "newClass",   dim, [newAIPSClass])
    if newFITSName:
        dim[0] = len(newFITSName)
        InfoList.PAlwaysPutString  (inInfo, "newFileName",dim, [newFITSName])
    # Rename
    Obit.UVRename (inUV.me, err.me)
    if err.isErr:
        OErr.printErrMsg(err, "Error Renaming UV data")
    # end PRename

def PCopy (inUV, outUV, err):
    """ Make a deep copy of input object.

    Makes structure the same as inUV, copies data, tables
    inUV   = Python UV object to copy
    outUV  = Output Python UV object, must be defined
    err    = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if not PIsA(outUV):
        raise TypeError,"outUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    Obit.UVCopy (inUV.me, outUV.me, err.me)
    if err.isErr:
        OErr.printErrMsg(err, "Error copying UV data")
    # end PCopy

def PClone (inUV, outUV, err):
    """ Make a copy of a object but do not copy the actual data

    This is useful to create an UV similar to the input one.
    inUV   = Python UV object
    outUV  = Output Python UV object, must be defined
    err    = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if not PIsA(outUV):
        raise TypeError,"outUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    Obit.UVClone (inUV.me, outUV.me, err.me)
    if err.isErr:
        OErr.printErrMsg(err, "Error cloning UV data")
    # end PClone

def PNewUVTable (inUV, access, tabType, tabVer, err):
    """ Obsolete use PGetTable
    """
    return  PGetTable (inUV, access, tabType, tabVer, err)
# end  PNewUVTable

def PGetTable (inUV, access, tabType, tabVer, err, \
               numOrb=0, numPCal=3, numIF=1, numPol=1, \
               numTerm=0, numChan=1, numTones=1, numBand=1, \
               numTabs=1, npoly=1, numCoef=5,
               maxis1=2, maxis2=1, maxis3=1, maxis4=1, maxis5=1):
    """ Return (create)the specified associated table
    
    Specific table types are recognized and the appropriate constructor
    called, these may have additional parameters.  This allows creating
    new tables of the appropriate type.
    returns Python Obit Table
    inUV      = Python UV object
    access    = access code 1=READONLY, 2=WRITEONLY, 3=READWRITE
    tabType   = Table type, e.g. "AIPS AN"
    tabVer    = table version, if > 0 on input that table returned,
                if 0 on input, the highest version is used.
    err       = Python Obit Error/message stack
    Optional parameters, values only used if table created
    numOrb    = Number of orbital parameters (AN)
    numPCal   = Number of polarization parameters (AN)
    numIF     = Number of IFs (FQ, SN, CL, BP, BL, TY, CQ)
    numPol    = Number of Stokes' (SN, CL, BP, BL, PC, TY, GC, MC, IM)
    numTerm   = Number of terms in model polynomial (CL)
    numChan   = Number of spectral channels (BP)
    numTomes  = Number of Phase cal tones (PC)
    numTabs   = Number of ??? (GC)
    numCoef   = Number of polynomial coefficents (NI)
    numBand   = Number  Bands(?) (IM, GC)
    npoly     = number of polynomial terms (IM)
    maxis1-5  = Dimension of axes of IDI data matrix
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    outTable    = Table.Table("None")
    id  = Obit.UVCastData(inUV.me)
    outTab = None
    if tabType=="AIPS AN":
        outTab = Obit.TableAN(id, [tabVer], access, tabType, numOrb, numPCal, err.me)
    elif tabType=="AIPS SU":
        outTab = Obit.TableSU(id, [tabVer], access, tabType, numIF, err.me)
    elif tabType=="AIPS NX":
        outTab = Obit.TableNX(id, [tabVer], access, tabType, err.me)
    elif tabType=="AIPS FQ":
        outTab = Obit.TableFQ(id, [tabVer], access, tabType, numIF, err.me)
    elif tabType=="AIPS FG":
        outTab = Obit.TableFG(id, [tabVer], access, tabType, err.me)
    elif tabType=="AIPS SN":
        outTab = Obit.TableSN(id, [tabVer], access, tabType, numPol, numIF, err.me)
    elif tabType=="AIPS CL":
        outTab = Obit.TableCL(id, [tabVer], access, tabType,
                              numPol, numIF, numTerm, err.me)
    elif tabType=="AIPS BP":
        outTab = Obit.TableBP(id, [tabVer], access, \
                              tabType, numPol, numIF, numChan, err.me)
    elif tabType=="AIPS BL":
        outTab = Obit.TableBL(id, [tabVer], access, tabType, numPol, numIF, err.me)
    elif tabType=="AIPS PC":
        outTab = Obit.TablePC(id, [tabVer], access, tabType, numPol, numTones, err.me)
    elif tabType=="AIPS WX":
        outTab = Obit.TableWX(id, [tabVer], access, tabType, err.me)
    elif tabType=="AIPS TY":
        outTab = Obit.TableTY(id, [tabVer], access, tabType, numPol, numIF, err.me)
    elif tabType=="AIPS GC":
        outTab = Obit.TableGC(id, [tabVer], access, tabType, \
                              numBand, numPol, numTabs, err.me)
    elif tabType=="AIPS MC":
        outTab = Obit.TableMC(id, [tabVer], access, tabType, numPol, err.me)
    elif tabType=="AIPS NI":
        outTab = Obit.TableNI(id, [tabVer], access, tabType, numCoef, err.me)
    elif tabType=="AIPS PS":
        outTab = Obit.TablePS(id, [tabVer], access, tabType, err.me)
    elif tabType=="AIPS CQ":
        outTab = Obit.TableCQ(id, [tabVer], access, tabType, numIF, err.me)
    elif tabType=="AIPS IM":
        outTab = Obit.TableIM(id, [tabVer], access, tabType, \
                              numBand, numPol, npoly, err.me)
    elif tabType=="AIPS CT":
        outTab = Obit.TableCT(id, [tabVer], access, tabType, err.me)
    elif tabType=="AIPS OB":
        outTab = Obit.TableOB(id, [tabVer], access, tabType, err.me)
    elif tabType=="AIPS OF":
        outTab = Obit.TableOF(id, [tabVer], access, tabType, err.me)
        # IDI tables
    elif tabType=="IDI_ANTENNA":
        outTab = Obit.TableIDI_ANTENNA(id, [tabVer], access, tabType,
                                       numIF, numPCal, err.me)
    elif tabType=="IDI_ARRAY_GEOMETRY":
        outTab = Obit.TableIDI_ARRAY_GEOMETRY(id, [tabVer], access, tabType,
                                              numIF, numOrb, err.me)
    elif tabType=="IDI_FREQUENCY":
        outTab = Obit.TableIDI_FREQUENCY(id, [tabVer], access, tabType,
                                         numIF, err.me)
    elif tabType=="IDI_SOURCE":
        outTab = Obit.TableIDI_SOURCE(id, [tabVer], access, tabType,
                                      numIF, err.me)
    elif tabType=="IDI_UV_DATA":
        outTab = Obit.TableIDI_UV_DATA(id, [tabVer], access, tabType,
                                       numIF, maxis1, maxis2, maxis3, maxis4, maxis5,
                                       err.me)
    else:  # Generic
        ret = Obit.newUVTable (inUV.me, access, tabType, [tabVer], err.me)
        # Table and version returned in a list
        outTab = ret[0]
    # Did it work? error may not be set
    if (outTab.__class__!=str) or (string.find(outTab,"_ObitTable_p") <0):
        OErr.printErrMsg(err, "Error getting UV data Table")
        raise RuntimeError,"Failed to extract "+tabType+" table from "+PGetName(inUV)
    # Error?
    if err.isErr:
        OErr.printErrMsg(err, "Error getting UV data Table")
    # Test validity of outTab
    if not Obit.TableIsA(outTab):
        OErr.printErrMsg(err, "Error getting UV data Table")
        raise RuntimeError,"Failed to extract "+tabType+" table from "+PGetName(inUV)
    # Open and close to fully instantiate - should exist
    outTable.me = outTab
    Table.PFullInstantiate (outTable, access, err)
    # Make sure that it worked - the output should be a table
    if not Table.PIsA(outTable):
        raise RuntimeError,"Failed to extract "+tabType+" table from "+PGetName(inUV)
    return outTable
    # end PGetTable


def POpen (inUV, access, err):
    """ Open an image persistent (disk) form

    inUV   = Python UV object
    access    = access 1=READONLY, 2=WRITEONLY, 3=READWRITE
    err       = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    ret = Obit.UVOpen(inUV.me, access, err.me)
    if err.isErr:
        OErr.printErrMsg(err, "Error Opening UV data")
    return ret
    # end POpen

def PDirty (inUV):
    """ Mark UV as needing a header update to disk file

    inUV     = Python UV object
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    #
    Obit.UVDirty (inUV.me)
    # end PDirty

def PClose (inUV, err):
    """ Close an image  persistent (disk) form

    inUV   = Python UV object
    err       = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    ret = Obit.UVClose (inUV.me, err.me)
    if err.isErr:
        OErr.printErrMsg(err, "Error Closing UV data")
    return ret
    # end PClose

def PZapTable (inUV, tabType, tabVer, err):
    """ Destroy specified table

    inUV      = Python UV object
    tabType   = Table type, e.g. "AIPS AN"
    tabVer    = table version, integer
    err       = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    ret = Obit.UVZapTable(inUV.me, tabType, tabVer, err.me)
    if err.isErr:
        OErr.printErrMsg(err, "Error Zapping UV data Table")
    # Update header
    inUV.UpdateTables(err)
    return ret
    # end PZapTable

def PCopyTables (inUV, outUV, exclude, include, err):
    """ Copy Tabeles from one image to another

    inUV      = Python UV object
    outUV     = Output Python UV object, must be defined
    exclude   = list of table types to exclude (list of strings)
                has priority
    include   = list of table types to include (list of strings)
    err       = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if not PIsA(outUV):
        raise TypeError,"outUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    ret = Obit.UVCopyTables  (inUV.me, outUV.me, exclude, include, err.me)
    if err.isErr:
        OErr.printErrMsg(err, "Error copying Tables")
    return ret
    # end PCopyTables

def PUpdateTables (inUV, err):
    """ Update any disk resident structures about the current tables

    inUV      = Python UV object
    err       = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    ret = Obit.UVUpdateTables (inUV.me, err.me)
    if err.isErr:
        OErr.printErrMsg(err, "Error updating Tables info")
    return ret
    # end PUpdateTables

def PFullInstantiate (inUV, access, err):
    """ Fully instantiate an UV by opening and closing

    return 0 on success, else failure
    inUV   = Python UV object
    access    = access code 1=READONLY, 2=WRITEONLY, 3=READWRITE
    err       = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    ret = Obit.UVfullInstantiate (inUV.me, access, err.me)
    if err.isErr:
        OErr.printErrMsg(err, "Error verifying UV data")
    return ret
    # end PfullInstantiate

def PGetList (inUV):
    """ Return the member InfoList

    returns InfoList
    inUV   = Python UV object
    """
    ################################################################
     # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    #
    out    = InfoList.InfoList()
    out.me = Obit.InfoListUnref(out.me)
    out.me = Obit.UVGetList(inUV.me)
    return out
    # end PGetList

def PGetTableList (inUV):
    """ Return the member tableList

    returns tableList
    inUV   = Python UV object
    """
    ################################################################
     # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    #
    out    = TableList.TableList(PGetName(inUV))
    out.me = Obit.TableListUnref(out.me)
    out.me = Obit.UVGetTableList(inUV.me)
    return out
    # end PGetTableList


def PHeader (inUV, err):
    """ Print data descriptor

    inUV      = Python Obit UV object
    err       = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV data"
    #
    # Fully instantiate
    PFullInstantiate (inUV, READONLY, err)
    # File info
    if inUV.FileType=="AIPS":
        print "AIPS UV Data Name: %12s Class: %6s seq: %8d disk: %4d" % \
              (inUV.Aname, inUV.Aclass, inUV.Aseq, inUV.Disk)
    elif inUV.FileType=="FITS":
        print "FITS UV Data Disk: %5d File Name: %s " % \
              (inUV.Disk, inUV.FileName)
    # print in UVDesc
    UVDesc.PHeader(inUV.Desc)
    # Tables
    TL = inUV.TableList
    Tlist = TableList.PGetList(TL, err)
    Tdict = {}
    # Once to find everything
    for item in Tlist:
        Tdict[item[1]] = item[0]
    # Again to get Max
    for item in Tlist:
        count = max (Tdict[item[1]], item[0])
        Tdict[item[1]] = count
    for item,count in Tdict.items():
        print "Maximum version number of %s tables is %d " % \
              (item, count)
    # end PHeader
    

def PGetDesc (inUV):
    """ Return the member UVDesc

    returns UVDesc as a Python Dictionary
    inUV   = Python UV object
    """
    ################################################################
     # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    #
    out    = UVDesc.UVDesc("None")
    out.me = Obit.UVGetDesc(inUV.me)
    return out
    # end PGetDesc

def PUpdateDesc (inUV, err, Desc=None):
    """ Update external representation of descriptor

    inUV   = Python UV object
    err    = Python Obit Error/message stack
    Desc   = UV descriptor, if None then use current descriptor
             Contents can be accessed throuth the Dict member
   """
    ################################################################
     # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    #
    # if Desc=None make copy of current contents
    if Desc == None:
        d = inUV.Desc.Dict
    else:
        d = Desc.Dict
    # Open for write
    inUV.Open(READWRITE,err) # Open
    inUV.Desc.Dict = d      # Update header
    Obit.UVDirty(inUV.me)   # force update
    inUV.Close(err)         # Close to update
    # end PUpdateDesc

def PGetVisBuf (inUV):
    return Obit.UVGetVisBuf(inUV.me)

def PGetHighVer (inUV, tabType):
    """ Get highest version number of a specified Table

    returns highest tabType version number, 0 if none.
    inUV   = Python UV object
    tabType   = Table type, e.g. "OTFSoln"
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    #
    return Obit.UVGetHighVer(inUV.me, tabType)
    # end PGetHighVer

def PGetSubA (inUV, err):
    """ Get Subarray information

    returns 0 on success, else 1
    inUV   = Python UV object
    err       = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    ret = Obit.UVGetSubA(inUV.me, err.me)
    if err.isErr:
        OErr.printErrMsg(err, "Error getting subarray info")
    # end PGetSubA 

def PGetFreq (inUV, err):
    """ Get Frequency information

    inUV   = Python UV object
    err       = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    ret = Obit.UVGetSubA(inUV.me, err.me)
    if err.isErr:
        OErr.printErrMsg(err, "Error getting frequency info")
    return ret
    # end PGetFreq


def PIsScratch (inUV):
    """ Tells if UV is a scratch object

    return true, false (1,0)
    inUV   = Python UV object
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    #
    return Obit.UVisScratch(inUV.me)
    # end PIsScratch

def PIsA (inUV):
    """ Tells if input really a Python Obit UV

    return true, false (1,0)
    inUV   = Python UV object
    """
    ################################################################
     # Checks
    if inUV.__class__ != UV:
        return 0
    return Obit.UVIsA(inUV.me)
    # end PIsA

def PGetName (inUV):
    """ Tells UV object name (label)

    returns name as character string
    inUV   = Python UV object
    """
    ################################################################
     # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    #
    return Obit.UVGetName(inUV.me)
    # end PGetName

#----------------------  UVUtil routines    ------------------------

def PUtilUVWExtrema (inUV, err):
    """ Get UV coverage information

    returns array [0]=maximum baseline length (in U,V), [1] = maximum W
    inUV   = Python UV object
    err    = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    val = [0.0, 0.0] # create array
    Obit.UVUtil(inUV.me, err.me, val)
    if err.isErr:
        OErr.printErrMsg(err, "Error getting  uv coverage info")
    return val
    # end PUTilUVWExtrema

def PUtilCopyZero (inUV, scratch, outUV, err):
    """ Copy a UV data set replacing data by zero, weight 1

    returns UV data object
    inUV   = Python UV object to copy
    scratch= True if this is to be a scratch file (same type as inUV)
    outUV  = Predefined UV data if scratch is False
             ignored if scratch True.
    err    = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if ((not scratch) and (not PIsA(outUV))):
        raise TypeError,"outUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    # Create output for scratch
    if scratch:
        outUV = UV("None")
    outUV.me = Obit.UVUtilCopyZero(inUV.me, scratch, outUV.me, err.me)
    if err.isErr:
        OErr.printErrMsg(err, "Error copying UV data to zeroed UV data")
    return outUV
    # end PUTilCopyZero

def PUtilVisDivide (in1UV, in2UV, outUV, err):
    """ Divides the visibilites in in1UV by those in in2UV

    in1UV   = Numerator  Python UV object
    in2UV   = Denominator Python UV object
    outUV   = Output Python UV object
    err       = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(in1UV):
        raise TypeError,"in1UV MUST be a Python Obit UV"
    if not PIsA(in2UV):
        raise TypeError,"in2UV MUST be a Python Obit UV"
    if not PIsA(outUV):
        raise TypeError,"outUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    Obit.UVUtilVisDivide(in1UV.me, in2UV.me, outUV.me, err.me)
    if err.isErr:
        OErr.printErrMsg(err, "Error dividing UV data")
    # end PUtilVisDivide

def PUtilVisCompare (in1UV, in2UV, err):
    """ Compares the visibilites in in1UV with those in in2UV

    returns RMS real, imaginary parts/amplitude
    in1UV   = Numerator  Python UV object
    in2UV   = Denominator Python UV object
    err       = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(in1UV):
        raise TypeError,"in1UV MUST be a Python Obit UV"
    if not PIsA(in2UV):
        raise TypeError,"in2UV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    ret = Obit.UVUtilVisCompare(in1UV.me, in2UV.me, err.me)
    if err.isErr:
        OErr.printErrMsg(err, "Error comparing UV data")
    return ret
    # end PUtilVisCompare

def PUtilIndex (inUV, err, maxScan=None, maxGap=None):
    """ Indexes a uv data

    inUV    = Python UV object to index
    err     = Python Obit Error/message stack
    maxScan = max. scan length in min. [def. long]
    maxGap  = max. scan gap in min. [def. long]
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    # Max scan time specified?
    if maxScan !=None:
        inInfo = PGetList(inUV)    # 
        dim = [1,1,1,1,1]
        InfoList.PAlwaysPutFloat (inInfo, "maxScan", dim, [maxScan])
    # Max scan gap specified?
    if maxGap !=None:
        inInfo = PGetList(inUV)    # 
        dim = [1,1,1,1,1]
        InfoList.PAlwaysPutFloat (inInfo, "maxGap", dim, [maxGap])
    # Index
    Obit.UVUtilIndex(inUV.me, err.me)
    if err.isErr:
        OErr.printErrMsg(err, "Error indexing UV data")
    # end PUtilIndex

def PUtilAvgF (inUV, outUV, err, scratch=False, 
               NumChAvg=0, doAvgAll=False, ChanSel=None):
    """ Average A UV data set in Frequency

    returns Averaged UV data object
    inUV   = Python UV object to copy
             Any selection editing and calibration applied before average.
    outUV  = Predefined UV data if scratch is False, ignored if
             scratch is True.
    err    = Python Obit Error/message stack
    scratch  = True if this is to be a scratch file (same type as inUV)
    NumChAvg = Number of channels to average, [def.0  = all]
    doAvgAll = If TRUE then average all channels and IF.
    ChanSel  =  Groups of channels to consider (relative to channels &
                IFs selected by BChan, EChan, BIF, EIF)
                (start, end, increment, IF) as array of tuples
                where start and end at the beginning and ending
                channel numbers (1-rel) of the group to be included,
                increment is the increment between selected channels
                and IF is the IF number (1-rel)
                default increment is 1, IF=0 means all IF.
                Default is all channels in each IF.
                Example [(3,14,1,0),(25,30,1,0)] averages channels
                3 through 14 and 25 through 30 in each IF.
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if ((not scratch) and (not PIsA(outUV))):
        raise TypeError,"outUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    # Save parameters
    dim = [1,1,1,1,1]
    inInfo = PGetList(inUV)    # 
    InfoList.PAlwaysPutBoolean (inInfo, "doAvgAll",  dim, [doAvgAll])
    InfoList.PAlwaysPutInt     (inInfo, "NumChAvg",  dim, [NumChAvg])
    if ChanSel:
        temp=[]
        for x in ChanSel:
            temp.append(x[0]); temp.append(x[1]); 
            temp.append(x[2]); temp.append(x[3]);
        temp.append(0); temp.append(0); 
        temp.append(0); temp.append(0);
        dim[0] = 4; dim[1] = len(ChanSel)+1
        InfoList.PAlwaysPutInt  (inInfo, "ChanSel",  dim, temp)

    # Create output for scratch
    if scratch:
        outUV = UV("None")
    outUV.me = Obit.UVUtilAvgF(inUV.me, scratch, outUV.me, err.me)
    if err.isErr:
        OErr.printErrMsg(err, "Error averaging UV data")
    return outUV
    # end PUtilAvgF

def PUtilAvgT (inUV, outUV, err, scratch=False, timeAvg=1.0):
    """ Average A UV data set in Time

    returns Averaged UV data object
    inUV   = Python UV object to copy
             Any selection editing and calibration applied before average.
    outUV  = Predefined UV data if scratch is False, ignored if
             scratch is True.
    err    = Python Obit Error/message stack
    scratch  = True if this is to be a scratch file (same type as inUV)
    timeAvg  = Averaging time in min
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if ((not scratch) and (not PIsA(outUV))):
        raise TypeError,"outUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    # Save parameter
    dim = [1,1,1,1,1]
    inInfo = PGetList(inUV)    # 
    InfoList.PAlwaysPutFloat (inInfo, "timeAvg",  dim, [timeAvg])
    
    # Create output for scratch
    if scratch:
        outUV = UV("None")
    outUV.me = Obit.UVUtilAvgT(inUV.me, scratch, outUV.me, err.me)
    if err.isErr:
        OErr.printErrMsg(err, "Error averaging UV data")
    return outUV
    # end PUtilAvgT

def PUtilCount (inUV, err, timeInt=1440.0):
    """ Count data values by interval in a UV dataset

    Each new source starts a new interval
    returns a dist with entries:
    numTime  = Number of time intervals
    numCorr  = Number of Correlations per vis
    Count    = Number of good correlation/visibilities
    Bad      = Number of flagged correlation/visibilities
    Source   = Source ID per interval (or 0 if no source ID)
    LST      = Average LST (days) per interval
    
    inUV   = Python UV object to copy
             Any selection editing and calibration applied before average.
    err    = Python Obit Error/message stack
    timeInt  = interval  in min (max 500 intervals)
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    #
    list    = InfoList.InfoList()
    # Get values in info List
    list.me = Obit.UVUtilCount (inUV.me, timeInt, err.me)
    if err.isErr:
        OErr.printErrMsg(err, "Error counting UV data")
    # Save values
    out = {}
    dlist = InfoList.PGetDict(list)  # List as dict
    for x in dlist:
        if dlist[x][1][0]==1:  # Scalar or array
            out[x] = dlist[x][2][0]
        else:
            out[x] = dlist[x][2]
    del list, dlist
    
    return out
    # end PUtilCount

def PEditTD (inUV, outUV, err):
    """ Time-domain editing of UV data - produces FG table

    Fill flagging table with clipping by RMS values of the real and imaginary
    parts.  All correlations are clipped on each baseline if the RMS is
    larger than  the maximum.  The clipping is done independently in
    each time interval defined by timeAvg. 
       The clipping level is given by MIN (A, MAX (B,C)) where:
    A = sqrt (maxRMS[0]**2 + (avg_amp * maxRMS[1])**2)
       and avg_amp is the average amplitude on each baseline.
    B = median RMS + 3 * sigma of the RMS distribution.
    C = level corresponding to 3% of the data.
       All data on a given baseline/correlator are flagged if the RMS
    exceeds the limit.  If a fraction of bad baselines on any correlator
    exceeds maxBad, then all data to that correlator is flagged.  In
    addition, if the offending correlator is a parallel hand correlator
    then any corresponding cross hand correlations are also flagged.
    Flagging entries are written into FG table flagTab.
    Control parameters on inUV info member
      "flagTab" OBIT_int    (1,1,1) FG table version number [ def. 1]
      "timeAvg" OBIT_float  (1,1,1) Time interval over which to determine 
                data to be flagged (days) [def = 1 min.]
                NB: this should be at least 2 integrations.
      "maxRMS"  OBIT_float (2,1,1) Maximum RMS allowed, constant plus 
                amplitude coefficient. 
      "maxBad"  OBIT_float (1,1,1) Fraction of allowed flagged baselines 
                [default 0.25]

    inUV   = Python UV object to clip/flag
    outUV  = UV data onto which the FG table is to be attached.
             May be the same as inUV.
    err    = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if ((not scratch) and (not PIsA(outUV))):
        raise TypeError,"outUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    # Create output for scratch
    if scratch:
        outUV = UV("None")
        outUV.me = Obit.UVEditTD(inUV.me, outUV.me, err.me)
        if err.isErr:
            OErr.printErrMsg(err, "Error in Time domain edit")
    return outUV
    # end PEditTD

def PEditFD (inUV, outUV, err):
    """ Frequency-domain editing of UV data - produces FG table

    Editing is done independently for each visibility channel.  
    First clipping is done on correlator and Vpol amplitudes.  
    Following this, an average and RMS is determined for each channel 
    in each timeAvg period and a spectral baseline is established
    for the average values, either using a  median window filter (FDwidMW>0) 
    or a linear baseline fit (FDwidMW<=0) to specified channels.  
    Channels with excessive RMSes or residual amplitudes are flagged.
    Flagging is done by entering the offending data in FG table flagTab
    on outUV.
    Control parameters on inUV info member
      "flagTab" OBIT_int    (1,1,1) FG table version number [ def. 1]
      "timeAvg" OBIT_float  (1,1,1) Time interval over which to average 
                data to be flagged (days) [def = 1 min.]
      "FDmaxAmp"  OBIT_float (1,1,1) Maximum average amplitude allowed in the
                spectrum before fitting.  Any channel exceeding this is
                flagged in advance of the  baseline fitting or median
                filtering,. default = infinite 
      "FDmaxV"  OBIT_float (1,1,1) Maximum average amplitude allowed in V
                polarization; any channel exceeding this is flagged in
                advance of the  baseline fitting or median filtering, 
                Calculates V from difference in amplitudes.   
                default = infinite 
      "FDwidMW" OBIT_int (1,1,1) If > 0 the width of the median window in channels. 
                An odd number (5) is recommended,  default or 0 => linear baseline
      "FDmaxRMS" OBIT_float (2,1,1) Flag all channels having RMS 
                values > maxRMS[0] of the channel median sigma.[default = 6.]
                plus maxRMS[1] (default 0.1) of the channel average in quadrature
      "FDmaxRes" OBIT_float (1,1,1) Max. residual flux in sigma allowed for 
                channels outside the baseline fitting regions.  
                default = 6.
      "FDmaxResBL"  OBIT_float (1,1,1) Max. residual flux in sigma allowed for 
                channels within the baseline fitting regions. 
                Default = FDmaxRes
      "FDbaseSel"  OBIT_int (4,*,1) Channel selection to define spectral baseline 
                Used only for linear baseline fitting.
                Select groups of channels/IF(s) to fit as sets 
                of (Start,end,inc,IF), i.e., chanSel = 6,37,1,0, 
                92,123,1,0 for two regions applying to all IFs.  
                Channel and IF numbers 1 -rel
                The first group for which the end channel == 0 terminates the list
                Channel increments defaults to 1
                If the IF==0 then the group applies to all IF.
                Default is channels 2 => nchan-1 all IFs
    inUV   = Python UV object to flag
             Any prior selection and editing is applied.
    outUV  = UV data onto which the FG table is to be attached.
             May be the same as inUV.
    err    = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if ((not scratch) and (not PIsA(outUV))):
        raise TypeError,"outUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    # Create output for scratch
    if scratch:
        outUV = UV("None")
        outUV.me = Obit.UVEditFD(inUV.me, outUV.me, err.me)
        if err.isErr:
            OErr.printErrMsg(err, "Error in Frequency domain edit")
    return outUV
    # end PEditFD

def PEditStokes (inUV, outUV, err):
    """ Stokes editing of UV data, FG table out

       All data on a given baseline/correlator are flagged if the 
    amplitude of the datatype "FlagStok"  exceeds maxAmp.  
    If a fraction of bad baselines on any antenna/channel/IF exceeds 
    maxBad, then all data to that correlator is flagged.  
    Flagging entries are written into FG table flagTab.
    Results are unpredictable for uncalibrated data.
    Control parameters on info member of inUV:
      "flagStok" OBIT_string (1,1,1) Stokes value to clip (I, Q, U, V, R, L)
                 default = "V"
      "flagTab" OBIT_int    (1,1,1) FG table version number [ def. 1]
                NB: this should not also being used to flag the input data!
      "timeAvg" OBIT_float  (1,1,1) Time interval over which to determine
                data to be flagged (days) [def = 1 min.]
      "maxAmp"  OBIT_float (1,1,1) Maximum VPol allowed
      "maxBad"  OBIT_float (1,1,1) Fraction of allowed flagged baselines 
                to an antenna above which all baselines are flagged.
                [default 0.25]
 
    inUV   = Python UV object to clip/flag
    outUV  = UV data onto which the FG table is to be attached.
             May be the same as inUV.
    err    = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if ((not scratch) and (not PIsA(outUV))):
        raise TypeError,"outUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    # Create output for scratch
    if scratch:
        outUV = UV("None")
        outUV.me = Obit.UVEditStokes(inUV.me, outUV.me, err.me)
        if err.isErr:
            OErr.printErrMsg(err, "Error in clipping")
    return outUV
    # end PEditStokes

def PEditClip (inUV, scratch, outUV, err):
    """ Clip raw visibilities

    control parameters on inUV info member
       "maxAmp" OBIT_float  (1,1,1) Maximum allowed amplitude
       "oper"   OBIT_string (4,1,1) operation type:
            "flag" flag data with amplitudes in excess of maxAmp
            "clip" clip amplitudes at maxAmp and preserve phase
               default is "flag"
    returns UV data object
    inUV   = Python UV object to clip/flag
    scratch= True if this is to be a scratch file (same type as inUV)
    outUV  = Predefined UV data if scratch is False, may be inUV
             ignored if scratch True.
    err    = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if ((not scratch) and (not PIsA(outUV))):
        raise TypeError,"outUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    # Create output for scratch
    if scratch:
        outUV = UV("None")
        outUV.me = Obit.UVEditClip(inUV.me, scratch, outUV.me, err.me)
        if err.isErr:
            OErr.printErrMsg(err, "Error in clipping")
    return outUV
    # end PEditClip

def PEditClipStokes (inUV, scratch, outUV, err):
    """ Flag visibilities by Stokes

    Clip a uv data set.  Data with amplitudes of the selected stokes
    in excess of maxAmp are flagged.  Optionally all correlations associated
    may be flagged.  Stokes conversion as needed for test.
    Control parameters are on the inUV info member:
      "clipStok" OBIT_string (1,1,1) Stokes value to clip (I, Q, U, V, R, L)
                  default = "I"
       "flagAll"  Obit_bool   (1,1,1) if true, flag all associated correlations
                  default = True
       "maxAmp"   OBIT_float  (1,1,1) Maximum allowed amplitude
    returns UV data object
    inUV   = Python UV object to clip/flag
    scratch= True if this is to be a scratch file (same type as inUV)
    outUV  = Predefined UV data if scratch is False, may be inUV
             ignored if scratch True.
    err    = Python Obit Error/message stack
    """
    ################################################################
    # Checks
    if not PIsA(inUV):
        raise TypeError,"inUV MUST be a Python Obit UV"
    if ((not scratch) and (not PIsA(outUV))):
        raise TypeError,"outUV MUST be a Python Obit UV"
    if not OErr.OErrIsA(err):
        raise TypeError,"err MUST be an OErr"
    #
    # Create output for scratch
    if scratch:
        outUV = UV("None")
        outUV.me = Obit.UVEditClipStokes(inUV.me, scratch, outUV.me, err.me)
        if err.isErr:
            OErr.printErrMsg(err, "Error in clipping")
    return outUV
    # end PEditClipStokes

